from typing import Union


def request(response_type: Union[type, None]):
    """
    A decorator to register a request class with a specified response type.

    Args:
        response_type (Union[type, None]): The type of the response expected from the request, or None if no response is expected.

    Returns:
        type: The decorated class.

    Example:
        @request(ResponseType)
        class MyRequest:
            # Define the request class
            pass
    """

    def decorator(cls: type):
        cls._response = response_type
        cls._rmediator_decorated_request = True

        return cls

    return decorator



import pytest

from rmediator import Mediator
from rmediator.decorators import request, request_handler


class TestMediator:
    def test__mediator__valid_registration__successful(
        self, mediator, get_valid_request, get_valid_request_handler_first_type
    ):
        mediator.register_handler(
            get_valid_request, get_valid_request_handler_first_type()
        )

    def test__mediator__valid_registration_valid_handling__successful(
        self, mediator, get_valid_request, get_valid_request_handler_first_type
    ):
        mediator.register_handler(
            get_valid_request, get_valid_request_handler_first_type()
        )
        response = mediator.send(get_valid_request())

        assert type(response) is get_valid_request._response

    def test__mediator__valid_duplicate_handlers_registration__successful(
        self, mediator, get_valid_request, get_valid_request_handler_first_type
    ):
        mediator.register_handler(
            get_valid_request, get_valid_request_handler_first_type()
        )
        mediator.register_handler(
            get_valid_request, get_valid_request_handler_first_type()
        )

    def test__mediator__undecorated_request__raises_value_error(
        self, mediator, get_undecorated_request, get_valid_request_handler_first_type
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(
                get_undecorated_request, get_valid_request_handler_first_type()
            )

    def test__mediator__undecorated_request_handler__raises_value_error(
        self, mediator, get_valid_request, get_undecorated_request_handler
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(
                get_valid_request, get_undecorated_request_handler()
            )

    def test__mediator__invalid_types_given_to_register_funciton__raises_value_error(
        self, mediator
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(str, bool)

    def test__mediator__invalid_same_request_different_handlers_registration__raises_value_error(
        self,
        mediator,
        get_valid_request,
        get_valid_request_handler_first_type,
        get_valid_request_handler_second_type,
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(
                get_valid_request, get_valid_request_handler_first_type()
            )
            mediator.register_handler(
                get_valid_request, get_valid_request_handler_second_type()
            )

    def test__mediator__unregistered_request__raises_value_error(
        self,
        mediator,
        get_valid_request,
    ):
        with pytest.raises(ValueError):
            mediator.send(get_valid_request())

    def test__mediator__request_and_request_handler_response_type_mismatch__raises_value_error(
        self,
        mediator,
        get_valid_request,
        get_valid_request_handler_third_type,
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(
                get_valid_request, get_valid_request_handler_third_type()
            )

    def test__mediator__none_request_request_handler__raises_value_error(
        self,
        mediator,
        get_valid_request,
        get_none_request_request_handler,
    ):
        with pytest.raises(ValueError):
            mediator.register_handler(
                get_valid_request, get_none_request_request_handler()
            )


@pytest.fixture
def mediator():
    return Mediator()


@pytest.fixture
def get_response():
    class Response:
        pass

    return Response


@pytest.fixture
def get_valid_request(get_response):
    @request(get_response)
    class Request:
        pass

    return Request


@pytest.fixture
def get_undecorated_request(get_response):
    class Request:
        response = get_response

    return Request


@pytest.fixture
def get_valid_request_handler_first_type(get_valid_request, get_response):
    @request_handler(get_valid_request, get_response)
    class RequestHandler:
        pass

        def handle(self, request: get_valid_request) -> get_response:
            return get_response()

    return RequestHandler


@pytest.fixture
def get_undecorated_request_handler(get_valid_request, get_response):
    class RequestHandler:
        request = get_valid_request
        response = get_response

        def handle(self, request: get_valid_request) -> get_response:
            return get_response()

    return RequestHandler


@pytest.fixture
def get_none_request_request_handler(get_valid_request, get_response):
    class RequestHandler:
        request = None
        response = get_response

        def handle(self, request: None) -> get_response:
            return get_response()

    return RequestHandler


@pytest.fixture
def get_valid_request_handler_second_type(get_valid_request, get_response):
    @request_handler(get_valid_request, get_response)
    class RequestHandler:
        request = get_valid_request
        response = get_response

        def handle(self, request: get_valid_request) -> get_response:
            pass

    return RequestHandler


@pytest.fixture
def get_valid_request_handler_third_type(get_valid_request, get_response):
    @request_handler(get_valid_request, None)
    class RequestHandler:

        def handle(self, request: get_valid_request) -> None:
            pass

    return RequestHandler



from types import FunctionType

from rmediator.decorators import request


class TestRequest:
    def test__request__invalid_registration_no_response__raises_value_error(self):
        @request
        class InvalidRequest:
            pass

        assert type(InvalidRequest) is FunctionType
        assert type(InvalidRequest) is not type

    def test__request__valid_registration__successful(self):
        @request(str)
        class ValidRequest:
            pass

        assert ValidRequest._response is str
        assert hasattr(ValidRequest, "_rmediator_decorated_request") is True



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

TRequest = TypeVar("TRequest")
TResponse = TypeVar("TResponse")


class RequestHandler(Generic[TRequest, TResponse], ABC):
    _request = TRequest
    _response = TResponse

    @abstractmethod
    def handle(self, request: TRequest) -> TResponse:
        pass



from threading import Lock
from typing import Any, Dict

from rmediator.types import Request, RequestHandler


class SingletonMeta(type):
    """
    A thread-safe implementation of Singleton metaclass.

    This metaclass ensures that only one instance of a class is created and provides a global point of access to that instance.

    Attributes:
        _instances (dict): A dictionary to hold the single instance of each class.
        _lock (Lock): A lock object to ensure thread-safe instantiation of the singleton.
    """

    _instances = {}
    _lock: Lock = Lock()

    def __call__(cls, *args, **kwargs):
        """
        Controls the instantiation of the singleton instance.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            object: The single instance of the class.
        """
        with cls._lock:
            if cls not in cls._instances:
                instance = super().__call__(*args, **kwargs)
                cls._instances[cls] = instance
        return cls._instances[cls]


class Mediator(metaclass=SingletonMeta):
    """
    Mediator class to handle the registration and dispatching of requests to their corresponding handlers.

    Attributes:
        __request_handlers (Dict[type, RequestHandler]): A dictionary mapping request types to their handlers.
    """

    def __init__(self) -> None:
        """
        Initializes a new instance of the Mediator class.
        """
        self.__request_handlers: Dict[type, RequestHandler] = {}

    def send(self, request: Request) -> Any:
        """
        Sends a request to its corresponding handler and returns the handler's response.

        Args:
            request (Request): The request object to be handled.

        Returns:
            Any: The response from the handler.

        Raises:
            ValueError: If no handler has been registered for the type of the request.
        """
        handler = self.__request_handlers.get(type(request))
        if not handler:
            raise ValueError(
                f"Request cannot be handled; no handler has been registered for {type(request)}."
            )

        return handler.handle(request)

    def register_handler(
        self, request_type: type[Request], handler: RequestHandler
    ) -> None:
        """
        Registers a handler for a specific request type.

        Args:
            request_type (type[Request]): The type of the request to be handled.
            handler (RequestHandler): The handler to process the request.

        Raises:
            ValueError: If the handler's response type does not match the request's response type,
                        or if another handler for the same request type has already been registered.
        """
        self.__check_request_validity(request_type)
        self.__check_request_handler_validity(type(handler))

        if handler._response is not request_type._response:  # type: ignore
            raise ValueError(
                f"Handler cannot be registered; handler response {handler._response} does not match request response {request_type._response}."
            )

        if request_type in self.__request_handlers:
            old_handler = self.__request_handlers[request_type]

            if type(handler) is not type(old_handler):
                raise ValueError(
                    f"Handler cannot be registered; another handler for request type {request_type} has already been registered. Check {type(self.__request_handlers[request_type])}."
                )

        self.__request_handlers[request_type] = handler

    def __check_request_validity(self, request_type: type) -> None:
        """
        Checks if a request type has been decorated using @request.

        Args:
            request_type (type): The request type to be checked.

        Raises:
            ValueError: If the request type has not been decorated using @request.
        """
        if not hasattr(request_type, "_rmediator_decorated_request"):
            raise ValueError(
                f"Handler cannot be registered; the request {request_type} class has not been decorated using @request."
            )

    def __check_request_handler_validity(self, handler_type: type) -> None:
        """
        Checks if a request handler type has been decorated using @request_handler.

        Args:
            handler_type (type): The request handler type to be checked.

        Raises:
            ValueError: If the handler type has not been decorated using @request_handler.
        """
        if not hasattr(handler_type, "_rmediator_decorated_request_handler"):
            raise ValueError(
                f"Handler cannot be registered; the request handler {type(handler_type)} class has not been decorated using @request_handler."
            )



import inspect
from typing import Union


def request_handler(request_type: type, response_type: Union[type, None]):
    """
    A decorator to register a request handler class with specified request and response types.

    Args:
        request_type (type): The type of the request that the handler should process.
        response_type (Union[type, None]): The type of the response that the handler should return, or None if no response is expected.

    Returns:
        type: The decorated class.

    Raises:
        AttributeError: If the request type is None, if the class does not contain a 'handle' method,
                        or if the 'handle' method does not have exactly one parameter (excluding 'self').
        ValueError: If the type of the 'handle' method's parameter does not match the request type,
                    or if the return type of the 'handle' method does not match the response type.

    Example:
        @request_handler(RequestType, ResponseType)
        class MyHandler:
            def handle(self, request: RequestType) -> ResponseType:
                # Handle the request and return a response
                pass
    """

    def decorator(cls: type):
        if request_type is None:
            raise AttributeError(
                f"Request handler class cannot be registered; request type cannot be {None}."
            )

        if "handle" not in cls.__dict__:
            raise AttributeError(
                f"Request handler class cannot be registered; {cls} must contain method 'handle' that handles the request."
            )

        handle_method_signature = inspect.signature(cls.handle)
        if len(handle_method_signature.parameters) != 2:
            raise AttributeError(
                f"Request handler class cannot be registered; {cls} handle method must have exactly one parameter."
            )

        request_parameter = None
        for name in handle_method_signature.parameters.keys():
            if name == "self":
                continue

            request_parameter = name

        if cls.handle.__annotations__.get(request_parameter, None) != request_type:
            raise ValueError(
                f"Request handler class cannot be registered; {cls} handle method must have 'request' parameter of type {request_type}. In your method, {request_parameter} must have type {request_type}"
            )

        if cls.handle.__annotations__.get("return", None) != response_type:
            raise ValueError(
                f"Request handler class cannot be registered; {cls} handle method must have return type of {response_type}."
            )

        cls._request = request_type
        cls._response = response_type
        cls._rmediator_decorated_request_handler = True

        return cls

    return decorator



import pytest

from rmediator.decorators import request, request_handler


class TestRequestHandler:
    def test__request_handler__no_handle_method__raises_value_error(
        self, get_valid_request, get_response
    ):
        with pytest.raises(AttributeError):

            @request_handler(get_valid_request, get_response)
            class InvalidRequestHandler:
                pass

    def test__request_handler__request_type_mismatch__raises_value_error(
        self, get_valid_request, get_response
    ):
        with pytest.raises(AttributeError):

            @request_handler(get_valid_request, get_response)
            class InvalidRequestHandler:
                def handle(self, request: get_valid_request.response) -> None:
                    pass

    def test__request_handler__response_type_mismatch__raises_value_error(
        self, get_valid_request, get_response
    ):
        with pytest.raises(ValueError):

            @request_handler(get_valid_request, get_response)
            class InvalidRequestHandler:
                def handle(self, request: get_valid_request) -> get_valid_request:
                    pass

    def test__request_handler__none_type_request__raises_value_error(self):
        with pytest.raises(AttributeError):

            @request_handler(None, str)  # type: ignore
            class InvalidRequestHandler:
                def handle(self, request: None) -> str:
                    return ""

    def test__request_handler__valid_handler__successful(
        self, get_valid_request, get_response
    ):
        @request_handler(get_valid_request, get_response)
        class ValidRequestHandler:
            def handle(self, request: get_valid_request) -> get_response:
                pass

    def test__request_handler__valid_handler_different_request_parameter_name__successful(
        self, get_valid_request, get_response
    ):
        @request_handler(get_valid_request, get_response)
        class ValidRequestHandler:
            def handle(self, some_request: get_valid_request) -> get_response:
                pass

    def test__request_handler__invalid_handler_request_parameter_wrong_type__successful(
        self, get_valid_request, get_response
    ):
        with pytest.raises(ValueError):

            @request_handler(get_valid_request, get_response)
            class InvalidRequestHandler:
                def handle(self, some_request: str) -> get_response:
                    pass

    def test__request_handler__invalid_handler_multiple_parameters_in_handle__successful(
        self, get_valid_request, get_response
    ):
        with pytest.raises(AttributeError):

            @request_handler(get_valid_request, get_response)
            class InvalidRequestHandler:
                def handle(
                    self, request: get_valid_request, additonal: int
                ) -> get_response:
                    pass


@pytest.fixture
def get_valid_request(get_response):
    @request(get_response)
    class Request:
        pass

    return Request


@pytest.fixture
def get_response():
    class Response:
        pass

    return Response
